#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Virtual broker.

TODO: 
    - add flag to signal long positions only allowed (eg. no CFD)
"""

import numpy as np

class Broker():
    
    def __init__(self, broker_config, utils):
        self.leverage           = 1
        self.commission         = 0
        self.spread             = 0
        self.margin_available   = 0
        self.portfolio_balance  = 0
        self.pending_positions  = {}
        self.open_positions     = {}
        self.closed_positions   = {}
        self.cancelled_orders   = {}
        self.total_trades       = 0
        self.profitable_trades  = 0
        self.peak_value         = 0
        self.low_value          = 0
        self.max_drawdown       = 0
        self.home_currency      = 'AUD'
        self.NAV                = 0
        self.unrealised_PL      = 0
        self.utils              = utils
    
    
    def place_order(self, order_details):
        '''
            Place order with broker.
        '''
        instrument  = order_details["instrument"]
        size        = order_details["size"]
        order_price = order_details["order_price"]
        stop_loss  = order_details["stop_loss"]
        stop_distance = order_details['stop_distance']
        
        if stop_loss is None and stop_distance is not None:
            pip_value   = self.utils.get_pip_ratio(instrument)
            stop_loss  = order_price - np.sign(size)*stop_distance*pip_value
        
        order_no = self.total_trades + 1
        
        # Create new_position and add position to self.pending_positions
        new_position = order_details.copy()
        new_position['order_ID'] = order_no
        
        self.pending_positions[order_no] = new_position
        
        # Update trade tally
        self.total_trades += 1
    
    
    def open_position(self, order_no, candle, limit_price = None):
        ''' Opens position with broker. '''
        
        # Calculate margin requirements
        current_price   = candle.Open
        pip_value       = self.utils.get_pip_ratio(self.pending_positions[order_no]['instrument'])
        size            = self.pending_positions[order_no]['size']
        HCF             = self.pending_positions[order_no]['HCF']
        position_value  = abs(size) * current_price * HCF
        margin_required = self.calculate_margin(position_value)
        
        if size > 0:
            spread_cost = 0.5*self.spread * pip_value
        else:
            spread_cost = -0.5*self.spread * pip_value
        
        if margin_required < self.margin_available:
            # Fill order
            new_position = self.pending_positions[order_no]
            new_position['time_filled'] = candle.name
            if limit_price is None:
                new_position['entry_price'] = candle.Open + spread_cost
            else:
                new_position['entry_price'] = limit_price + spread_cost
            self.open_positions[order_no] = new_position
            
        else:
            self.cancelled_orders[order_no] = self.pending_positions[order_no]
    
    
    def update_positions(self, candle):
        ''' 
            Updates orders and open positions based on current candle. 
        '''
        # Tally for positions opened this update
        opened_positions = 0
        
        # Update pending positions
        closing_orders = []
        for order_no in self.pending_positions:
            if self.pending_positions[order_no]['order_time'] != candle.name:
                if self.pending_positions[order_no]['order_type'] == 'market':
                    # Market order type
                    self.open_position(order_no, candle)
                    opened_positions += 1
                
                elif self.pending_positions[order_no]['order_type'] == 'stop-limit':
                    # Stop-limit order type
                    # Check if order_stop_price has been reached yet
                    
                    if candle.Low < self.pending_positions[order_no]['order_stop_price'] < candle.High:
                        # order_stop_price has been reached, change order type to 'limit'
                        self.pending_positions[order_no]['order_type'] = 'limit'
                    
                # This is in a separate if statement, as stop-limit order may
                # eventually be changed to limit orders
                if self.pending_positions[order_no]['order_type'] == 'limit':
                    # Limit order type
                    if self.pending_positions[order_no]['size'] > 0:
                        if candle.Low < self.pending_positions[order_no]['order_limit_price']:
                            self.open_position(order_no, candle, 
                                               self.pending_positions[order_no]['order_limit_price'])
                            opened_positions += 1
                    else:
                        if candle.High > self.pending_positions[order_no]['order_limit_price']:
                            self.open_position(order_no, candle, 
                                               self.pending_positions[order_no]['order_limit_price'])
                            opened_positions += 1
                            
                            
            if self.pending_positions[order_no]['order_type'] == 'close':
                related_order = self.pending_positions[order_no]['related_orders']
                self.close_position(self.pending_positions[order_no]['instrument'],
                                    candle, 
                                    candle.Close,
                                    order_no = related_order
                                    )
                opened_positions += 1 # To remove from pending orders
                closing_orders.append(order_no)
                    
        # Remove position from pending positions
        if opened_positions > 0:
            # For orders that were opened
            for order_no in self.open_positions.keys():
                self.pending_positions.pop(order_no, 0)
            
            # For orders that were cancelled
            for order_no in self.cancelled_orders.keys():
                self.pending_positions.pop(order_no, 0)
            
            # For close orders
            for order_no in closing_orders:
                self.pending_positions.pop(order_no, 0)
        
        # Update trailing stops
        # For other methods, move the stop update to an external function
        # Can this be moved into the loop below?
        for order_no in self.open_positions:
            if self.open_positions[order_no]['stop_type'] == 'trailing_stop':
                # Trailing stop loss is enabled, check if price has moved SL
                
                if self.open_positions[order_no]['stop_distance'] is not None:
                    pip_value = self.utils.get_pip_ratio(self.open_positions[order_no]['instrument'])
                    pip_distance = self.open_positions[order_no]['stop_distance']
                    distance = pip_distance*pip_value
                    
                else:
                    distance = abs(self.open_positions[order_no]['entry_price'] - \
                                   self.open_positions[order_no]['stop_loss'])
                

                if self.open_positions[order_no]['size'] > 0:
                    # long position, stop loss only moves up
                    new_stop = candle.High - distance
                    if new_stop > self.open_positions[order_no]['stop_loss']:
                        self.open_positions[order_no]['stop_loss'] = new_stop
                    
                else:
                    # short position, stop loss only moves down
                    new_stop = candle.Low + distance
                    if new_stop < self.open_positions[order_no]['stop_loss']:
                        self.open_positions[order_no]['stop_loss'] = new_stop
        
        # Update self.open_positions
        open_position_orders = list(self.open_positions.keys())
        unrealised_PL  = 0        # Un-leveraged value
        for order_no in open_position_orders:
            if self.open_positions[order_no]['size'] > 0:
                if self.open_positions[order_no]['stop_loss'] is not None and \
                    candle.Low < self.open_positions[order_no]['stop_loss']:
                    # Stop loss hit
                    self.close_position(self.open_positions[order_no]['instrument'], 
                                        candle, 
                                        self.open_positions[order_no]['stop_loss'],
                                        order_no)
                elif self.open_positions[order_no]['take_profit'] is not None and \
                    candle.High > self.open_positions[order_no]['take_profit']:
                    # Take Profit hit
                    self.close_position(self.open_positions[order_no]['instrument'], 
                                        candle, 
                                        self.open_positions[order_no]['take_profit'],
                                        order_no)
                else:
                    # Position is still open, update value of holding
                    size        = self.open_positions[order_no]['size']
                    entry_price = self.open_positions[order_no]['entry_price']
                    price       = candle.Close
                    HCF         = self.open_positions[order_no]['HCF']
                    position_value = size*(price - entry_price)*HCF
                    unrealised_PL += position_value
            
            else:
                if self.open_positions[order_no]['stop_loss'] is not None and \
                    candle.High > self.open_positions[order_no]['stop_loss']:
                    self.close_position(self.open_positions[order_no]['instrument'], 
                                        candle, 
                                        self.open_positions[order_no]['stop_loss'],
                                        order_no)
                elif self.open_positions[order_no]['take_profit'] is not None and \
                    candle.Low < self.open_positions[order_no]['take_profit']:
                    self.close_position(self.open_positions[order_no]['instrument'], 
                                        candle, 
                                        self.open_positions[order_no]['take_profit'],
                                        order_no)
                else:
                    size        = self.open_positions[order_no]['size']
                    entry_price = self.open_positions[order_no]['entry_price']
                    price       = candle.Close
                    HCF         = self.open_positions[order_no]['HCF']
                    position_value = size*(price - entry_price)*HCF
                    unrealised_PL += position_value
        
        # Update margin available
        self.update_margin(candle.Close)
        
        # Update unrealised P/L
        self.unrealised_PL = unrealised_PL
        
        # Update open position value
        self.NAV = self.portfolio_balance + self.unrealised_PL
    
    
    def close_position(self, instrument, candle, exit_price,
                           order_no=None):
        ''' Closes position. '''
        if order_no is not None:
            # single order specified to close
            self.close_trade(order_no=order_no, 
                             candle=candle, 
                             exit_price=exit_price)
        else:
            # Close all positions for instrument
            for order_no in self.open_positions.copy():
                if self.open_positions[order_no]['instrument'] == instrument:
                    self.close_trade(order_no=order_no, 
                                     candle=candle, 
                                     exit_price=exit_price)
    
    
    def close_trade(self, candle=None, exit_price=None,
                           order_no=None):
        ''' Closes trade by order number. '''
        entry_price = self.open_positions[order_no]['entry_price']
        size        = self.open_positions[order_no]['size']
        HCF         = self.open_positions[order_no]['HCF']
        
        # Update portfolio with profit/loss
        gross_PL    = size*(exit_price - entry_price)*HCF
        open_trade_value    = abs(size)*entry_price*HCF
        close_trade_value   = abs(size)*exit_price*HCF
        commission  = (self.commission/100) * (open_trade_value + close_trade_value)
        net_profit  = gross_PL - commission
        
        self.add_funds(net_profit)
        if net_profit > 0:
            self.profitable_trades += 1
        
        # Add trade to closed positions
        closed_position = self.open_positions[order_no].copy()
        closed_position['profit'] = net_profit
        closed_position['balance'] = self.portfolio_balance
        closed_position['exit_price'] = exit_price
        closed_position['exit_time'] = candle.name
        
        # Add to closed positions dictionary
        self.closed_positions[order_no] = closed_position
        
        # Remove position from open positions
        self.open_positions.pop(order_no, 0)
        
        # Update maximum drawdown
        self.update_MDD()
        
    
    def get_pending_orders(self, instrument = None):
        ''' Returns pending orders. '''
        return self.pending_positions
    
    def cancel_pending_order(self, order_id):
        self.pending_positions.pop(order_id, 0)
    
    def get_open_positions(self, instrument=None):
        ''' Returns the open positions in the account. '''
        return self.open_positions
    
    
    def add_funds(self, amount):
        ''' Add's funds to brokerage account. '''
        
        if self.portfolio_balance == 0:
            # If this is the initial deposit, set peak and low values for MDD
            self.peak_value = amount
            self.low_value = amount
            
        self.portfolio_balance  += amount
        # self.margin_available   += amount
    
    
    def get_balance(self):
        ''' Returns balance of account. '''
        return self.portfolio_balance
    
    
    def calculate_margin(self, position_value):
        ''' Calculates margin required to take a position. '''
        margin = position_value / self.leverage
        
        return margin
    
    
    def update_margin(self, close_price):
        ''' Updates margin available in account. '''
        
        margin_used = 0
        for order_no in self.open_positions:
            size            = self.open_positions[order_no]['size']
            HCF             = self.open_positions[order_no]['HCF']
            # HCF should be updated for current time.
            position_value  = abs(size) * close_price * HCF
            margin_required = self.calculate_margin(position_value)
            margin_used     += margin_required
        
        self.margin_available = self.portfolio_balance - margin_used


    def get_price(self, instrument, data, conversion_data, i):
        ''' Returns the price data dict. '''
        
        ask = data.Close[i]
        bid = data.Close[i]
        conversion_data = conversion_data.Close[i]
        
        if bid == conversion_data:
            negativeHCF = 1
            positiveHCF = 1
        else:
            negativeHCF = 1/conversion_data
            positiveHCF = 1/conversion_data
        
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF
                 }
        
        return price
    
    def update_MDD(self):
        ''' Function to calculate maximum portfolio drawdown '''
        balance     = self.portfolio_balance
        peak_value  = self.peak_value
        low_value   = self.low_value
        
        if balance > peak_value:
            self.peak_value = balance
            self.low_value = balance
            
        elif balance < low_value:
            self.low_value = balance
        
        MDD = 100*(low_value - peak_value)/peak_value
        
        if MDD < self.max_drawdown:
            self.max_drawdown = MDD
    
    def get_NAV(self):
        return self.NAV



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

AuotOptimise
-------------

Want to use a walk-forward optimisation approach. 

Plot output as heatmap: parameter grid coloured by objective function

Consider sample size, print as warning 

TODO: move into autotrader.py somehow...


"""

from autotrader.autotrader import AutoTrader
import yaml
import os
from getopt import getopt
import sys
from scipy.optimize import brute
import timeit
from ast import literal_eval


class Optimise():
    '''
    
    opt_params  = ['param1', 'param2', ...]
    bounds      = [  (),       (),     ...]
        
    '''
    
    def __init__(self):
        self.config_file    = None
        self.verbosity      = 1
        self.show_help      = None
        self.log            = False
        self.home_dir       = None
        self.opt_params     = None
        self.bounds         = None
        self.Ns             = 4
        self.objective      = 'profit + MDD'
    
    def read_yaml(self, file_path):
        '''Function to read and extract contents from .yaml file.'''
        with open(file_path, "r") as f:
            return yaml.safe_load(f)
    
    
    def helper_function(self, params, config_dict, opt_params, verbosity):
        '''
            Helper function for optimising strategy in AutoTrader.
            This function will parse the ordered params into the config dict.
            
        '''
        
        ''' ------------------------------------------------------------------ '''
        '''   Edit strategy parameters in config_dict using supplied params    '''
        ''' ------------------------------------------------------------------ '''
        for parameter in config_dict['STRATEGY']['PARAMETERS']:
            if parameter in opt_params:
                config_dict['STRATEGY']['PARAMETERS'][parameter] = params[opt_params.index(parameter)]
            else:
                continue
        
        ''' ------------------------------------------------------------------ '''
        '''           Run AutoTrader and evaluate objective function           '''
        ''' ------------------------------------------------------------------ '''
        at                  = AutoTrader()
        at.backtest         = True
        at.optimise         = True
        at.custom_config    = config_dict
        at.run()
        backtest_results    = at.backtest_results
        
        try:
            objective           = -backtest_results['all_trades']['profit_pc']  - \
                                  backtest_results['all_trades']['MDD']
        except:
            objective           = 100
                              
        print("Parameters/objective:", params, "/", objective)
        
        return objective
    
    
    def optimise(self):
        
        ''' ------------------------------------------------------------------ '''
        '''                          Unpack user options                       '''
        ''' ------------------------------------------------------------------ '''
        if self.home_dir is not None:
            home_dir            = self.home_dir
        else:
            home_dir            = os.getcwd()
        config_file_path    = os.path.join(home_dir, 'config', self.config_file)
        config_dict         = self.read_yaml(config_file_path + '.yaml')
        verbosity           = self.verbosity
        
        ''' ------------------------------------------------------------------ '''
        '''                      Define optimisation inputs                    '''
        ''' ------------------------------------------------------------------ '''
        if type(self.bounds) == str:
            full_tuple = literal_eval(self.bounds)
            bounds = [(x[0], x[-1]) for x in full_tuple]
        else:
            bounds = self.bounds

        if type(self.opt_params) == str:
            opt_params = self.opt_params.split(',')
        else:
            opt_params = self.opt_params
            
        my_args     = (config_dict, opt_params, verbosity)
        
        ''' ------------------------------------------------------------------ '''
        '''                             Run Optimiser                          '''
        ''' ------------------------------------------------------------------ '''
        start = timeit.default_timer()
        result = brute(func         = self.helper_function, 
                       ranges       = bounds, 
                       args         = my_args, 
                       Ns           = self.Ns,
                       full_output  = True)
        stop = timeit.default_timer()
        
        ''' ------------------------------------------------------------------ '''
        '''      Delete historical data file after running optimisation        '''
        ''' ------------------------------------------------------------------ '''
        granularity             = config_dict["STRATEGY"]["INTERVAL"]
        pair                    = config_dict["WATCHLIST"][0]
        historical_data_name    = 'hist_{0}{1}.csv'.format(granularity, pair)
        historical_quote_data_name = 'hist_{0}{1}_quote.csv'.format(granularity, pair)
        historical_data_file_path = os.path.join(home_dir, 
                                                 'price_data',
                                                 historical_data_name)
        historical_quote_data_file_path = os.path.join(home_dir, 
                                                 'price_data',
                                                 historical_quote_data_name)
        os.remove(historical_data_file_path)
        os.remove(historical_quote_data_file_path)
        
        
        opt_params = result[0]
        opt_value = result[1]
        
        # TODO - use the below for heatmap plotting
        grid_points = result[2]
        grid_values = result[3]
        
        
        ''' ------------------------------------------------------------------ '''
        '''                            Print output                            '''
        ''' ------------------------------------------------------------------ '''
        print("\nOptimisation complete.")
        print('Time to run: {}s'.format(round((stop - start), 3)))
        print("Optimal parameters:")
        print(opt_params)
        print("Objective:")
        print(opt_value)
        
    
    def print_usage(self):
        print("Help is on its way! Check back again later.")
    
    def print_help(self, option):
        print("Help is on its way! Check back again later.")



short_options = "h:c:v:lo:b:"
long_options = ["help=", "config=", "verbosity=", "log", "optimise=", 
                "bounds="]

if __name__ == '__main__':
    
    # Instantiate Optimiser
    optimiser = Optimise()
    
    # Extract User Inputs
    options, r = getopt(sys.argv[1:], 
                        short_options, 
                        long_options
                        )
    
    # Default Options
    config_file     = None
    verbosity       = 0
    show_help       = None
    log             = False
    opt_params      = None
    bounds          = None
    
    for opt, arg in options:
        if opt in ('-c', '--config'):
            config_file = arg
            optimiser.config_file = arg
        elif opt in ('-v', '--verbose'):
            verbosity = arg
            optimiser.verbose = arg
        elif opt in ('-h', '--help'):
            show_help = arg
            optimiser.show_help = arg
        elif opt in ('-l', '--log'):
            log = True
            optimiser.log = True
        elif opt in ('-o', '--optimise'):
            opt_params = arg
            optimiser.opt_params = arg
        elif opt in ('-b', '--bounds'):
            opt_params = arg
            optimiser.bounds = arg
            

    if len(options) == 0:
        optimiser.print_usage()
        
    elif show_help is not None:
        optimiser.print_help(show_help)
        
    else:
        optimiser.optimise()


# -*- coding: utf-8 -*-
"""

Oanda API Wrapper
=================

"""

import v20
from autotrader.brokers.oanda import utils
import datetime
import pandas as pd
import numpy as np


class Oanda():
    def __init__(self, oanda_config, utils):
        ''' Create v20 context. '''
        API             = oanda_config["API"]
        ACCESS_TOKEN    = oanda_config["ACCESS_TOKEN"]
        port            = oanda_config["PORT"]
        self.ACCOUNT_ID = oanda_config["ACCOUNT_ID"]
        self.api        = v20.Context(hostname=API, 
                                      token=ACCESS_TOKEN, 
                                      port=port)
        
        STREAM_API      = "stream-fxpractice.oanda.com"
        self.stream     = v20.Context(hostname=STREAM_API, 
                                      token=ACCESS_TOKEN, 
                                      port=443)
        
        self.open_positions     = {}
        
    
    def get_price(self, pair):
        ''' Returns current price (bid+ask) and home conversion factors.'''
        response = self.api.pricing.get(accountID = self.ACCOUNT_ID, 
                                   instruments = pair
                                   )
        ask = response.body["prices"][0].closeoutAsk
        bid = response.body["prices"][0].closeoutBid
        negativeHCF = response.body["prices"][0].quoteHomeConversionFactors.negativeUnits
        positiveHCF = response.body["prices"][0].quoteHomeConversionFactors.positiveUnits
    
        price = {"ask": ask,
                 "bid": bid,
                 "negativeHCF": negativeHCF,
                 "positiveHCF": positiveHCF
                 }
    
        return price
    
    def get_pending_orders(self, instrument=None):
        ''' Get all pending orders in the account. '''
        response = self.api.order.list_pending(accountID = self.ACCOUNT_ID, 
                                          instrument=instrument)
        
        oanda_pending_orders = response.body['orders']
        pending_orders = {}
        
        for order in oanda_pending_orders:
            if order.type != 'TAKE_PROFIT' and order.type != 'STOP_LOSS':
                new_order = {}
                new_order['order_ID']           = order.id
                new_order['order_type']         = order.type
                new_order['order_stop_price']   = order.price
                new_order['order_limit_price']  = order.price
                new_order['direction']          = np.sign(order.units)
                new_order['order_time']         = order.createTime
                new_order['strategy']           = None
                new_order['instrument']         = order.instrument
                new_order['size']               = order.units
                new_order['order_price']        = order.price
                new_order['granularity']        = None
                new_order['take_profit']        = order.takeProfitOnFill.price
                new_order['take_distance']      = None
                new_order['stop_type']          = None
                new_order['stop_distance']      = None
                new_order['stop_loss']          = None
                new_order['related_orders']     = None
                
                if instrument is not None and order.instrument == instrument:
                    pending_orders[order.id] = new_order
                elif instrument is None:
                    pending_orders[order.id] = new_order
            
        return pending_orders
    
    def cancel_pending_order(self, order_id):
        ''' Cancels pending order by ID. '''
        self.api.order.cancel(accountID = self.ACCOUNT_ID, orderSpecifier=str(order_id))
    
    def get_open_positions(self, pair = None):
        ''' Gets the current positions open on the account. '''
        
        if pair is not None:
            
            response = self.api.position.get(accountID = self.ACCOUNT_ID, 
                                             instrument = pair)
            # print(response.body['position'])
        else:
            response = self.api.position.list_open(accountID = self.ACCOUNT_ID)
            # print(response.body['positions'][0])
        
        return response.body
    
    def place_order(self, order_details):
        '''
        Parses order_details dict and handles order.
        '''
        
        if order_details["order_type"] == 'market':
            response = self.place_market_order(order_details)
        elif order_details["order_type"] == 'stop-limit':
            response = self.place_stop_limit_order(order_details)
        elif order_details["order_type"] == 'limit':
            response = self.place_limit_order(order_details)
        else:
            print("Order type not recognised.")
            return
        
        # Check response
        output = self.check_response(response)
        
        return response
        
    def place_market_order(self, order_details):
        ''' Places market order. '''
        stop_loss_details = self.get_stop_loss_details(order_details)
        take_profit_details = self.get_take_profit_details(order_details)
        
        response = self.api.order.market(accountID = self.ACCOUNT_ID,
                                         instrument = order_details["instrument"],
                                         units = order_details["size"],
                                         takeProfitOnFill = take_profit_details,
                                         stopLossOnFill = stop_loss_details,
                                         )
        return response
    
    def place_stop_limit_order(self, order_details):
        '''
        Places MarketIfTouchedOrder with Oanda.
        https://developer.oanda.com/rest-live-v20/order-df/
        '''
        stop_loss_details = self.get_stop_loss_details(order_details)
        take_profit_details = self.get_take_profit_details(order_details)
        
        # Check and correct order stop price
        price = self.check_precision(order_details["instrument"], 
                                     order_details["order_stop_price"])
        
        # Need to test cases when no stop/take is provided (as None type)
        response = self.api.order.market_if_touched(accountID   = self.ACCOUNT_ID,
                                                    instrument  = order_details["instrument"],
                                                    units       = order_details["size"],
                                                    price       = str(price),
                                                    takeProfitOnFill = take_profit_details,
                                                    stopLossOnFill = stop_loss_details
                                                    )
        return response
    
    def place_limit_order(self, order_details):
        ''' (NOT YET IMPLEMENTED) PLaces limit order. '''
        return

    def get_stop_loss_details(self, order_details):
        ''' Constructs stop loss details dictionary. '''
        # https://developer.oanda.com/rest-live-v20/order-df/#OrderType
        
        if order_details["stop_type"] is not None:
            price = self.check_precision(order_details["instrument"], 
                                         order_details["stop_loss"])
            
            if order_details["stop_type"] == 'trailing':
                # Trailing stop loss order
                stop_loss_details = {"price": str(price),
                                     "type": "TRAILING_STOP_LOSS"}
            else:
                stop_loss_details = {"price": str(price)}
        else:
            stop_loss_details = None
        
        return stop_loss_details
    
    def get_take_profit_details(self, order_details):
        ''' Constructs take profit details dictionary. '''
        if order_details["take_profit"] is not None:
            price = self.check_precision(order_details["instrument"], 
                                         order_details["take_profit"])
            take_profit_details = {"price": str(price)}
        else:
            take_profit_details = None
        
        return take_profit_details

    def get_data(self, pair, period, interval):
        # print("Getting data for {}".format(pair))
        response    = self.api.instrument.candles(pair,
                                             granularity = interval,
                                             count = period,
                                             dailyAlignment = 0
                                             )
        
        data        = utils.response_to_df(response)
        
        return data
    
    
    def get_balance(self):
        ''' Returns account balance. '''
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        
        return response.body["account"].balance
    
    
    def get_summary(self):
        ''' Returns account summary. '''
        # response = self.api.account.get(accountID=self.ACCOUNT_ID)
        response = self.api.account.summary(accountID=self.ACCOUNT_ID)
        # print(response.body['account'])
        
        return response
    
    
    def close_position(self, instrument, long_units=None, short_units=None,
                       **dummy_inputs):
        ''' Closes all open positions on an instrument '''
        # Check if the position is long or short
        position    = self.get_positions(instrument)['position']
        if long_units is None:
            long_units  = position.long.units
        if short_units is None:
            short_units = position.short.units
        
        if long_units > 0:
            response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                               instrument=instrument, 
                                               longUnits="ALL")
            # # Check response
            # output = self.check_response(response)
        
        elif short_units > 0: 
            response = self.api.position.close(accountID=self.ACCOUNT_ID, 
                                               instrument=instrument,
                                               shortUnits="ALL")
            # # Check response
            # output = self.check_response(response)
        
        else:
            print("There is no current position with {} to close.".format(instrument))
            response = None
        
        return response
    
    
    def get_precision(self, pair):
        ''' Returns the allowable precision for a given pair '''
        response = self.api.account.instruments(accountID = self.ACCOUNT_ID, 
                                                instruments = pair)
        
        precision = response.body['instruments'][0].displayPrecision
        
        return precision

    
    def check_precision(self, pair, price):
        ''' Modify a price based on required ordering precision for pair. ''' 
        N               = self.get_precision(pair)
        corrected_price = round(price, N)
        
        return corrected_price
        
    
    def check_response(self, response):
        ''' Checks API response (currently only for placing orders) '''
        if response.status != 201:
            message = response.body['errorMessage']
        else:
            message = "Success."
            
        output = {'Status': response.status, 
                  'Message': message}
        
        return output
    
    def update_data(self, pair, granularity, data):
        ''' Attempts to construct the latest candle when there is a delay in the 
            api feed.
        '''
        
        granularity_details = self.deconstruct_granularity(granularity)
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        # TODO: make this a function of original granularity
        # Done, but now I think, maybe this should always be something like S5?
        small_granularity = self.get_reduced_granularity(granularity_details, 
                                                         25)
        
        # Get data equivalent of last candle's granularity
        time_now        = datetime.datetime.now()
        start_time      = time_now - datetime.timedelta(seconds = secs,
                                                        minutes = mins,
                                                        hours = hrs,
                                                        days = days)
        latest_data     = self.get_historical_data(pair, 
                                                   small_granularity, 
                                                   start_time.timestamp(), 
                                                   time_now.timestamp())
        
        # Get latest price data
        latest_close    = latest_data.Close.values[0]
        
        open_price      = data.Close.values[-1]
        close_price     = latest_close
        high_price      = max(latest_data.High.values)
        low_price       = min(latest_data.Low.values)
        last_time       = data.index[-1]
        stripped_time   = datetime.datetime.strptime(last_time.strftime("%Y-%m-%d %H:%M:%S%z"),
                                                      "%Y-%m-%d %H:%M:%S%z")
        new_time        = stripped_time + datetime.timedelta(seconds = secs,
                                                              minutes = mins,
                                                              hours = hrs,
                                                              days = days)
        
        new_candle      = pd.DataFrame({'Open'  : open_price, 
                                        'High'  : high_price,
                                        'Low'   : low_price,
                                        'Close' : close_price},
                                        index=[new_time])
        
        
        new_data        = pd.concat([data, new_candle])
        
        return new_data
    
    
    def get_historical_data(self, pair, interval, from_time, to_time):
        response        = self.api.instrument.candles(pair,
                                                      granularity = interval,
                                                      fromTime = from_time,
                                                      toTime = to_time
                                                      )
        
        data = utils.response_to_df(response)
        
        return data
    
    
    def deconstruct_granularity(self, granularity):
        ''' Returns a dict with the seconds, minutes, hours and days
            corresponding to a granularity. 
        '''
        
        # Get letter to determine timeframe (eg. M)
        letter = granularity[0]
        
        # Get timeframe multiple (eg. 15)
        if len(granularity) > 1:
            number = float(granularity[1:])
        else:
            number = 1
        
        
        if letter == 'S':
            seconds     = number
            minutes     = 0
            hours       = 0
            days        = 0
        
        elif letter == 'M':
            seconds     = 0
            minutes     = number
            hours       = 0
            days        = 0
            
        elif letter == 'H':
            seconds     = 0
            minutes     = 0
            hours       = number
            days        = 0
            
        else:
            seconds     = 0
            minutes     = 0
            hours       = 0
            days        = number
        
        granularity_details = {'seconds': seconds,
                               'minutes': minutes,
                               'hours': hours,
                               'days': days}
        
        return granularity_details

    def get_reduced_granularity(self, granularity_details, fraction):
        '''Returns a candlestick granularity as a fraction of given granularity'''
        secs = granularity_details['seconds']
        mins = granularity_details['minutes']
        hrs  = granularity_details['hours']
        days = granularity_details['days']
        
        total_seconds = secs + 60*mins + 60*60*hrs + 60*60*24*days
        
        fractional_seconds = total_seconds/fraction
        
        seconds = fractional_seconds
        minutes = fractional_seconds/60
        hours = fractional_seconds/(60*60)
        days = fractional_seconds/(60*60*24)
        
        if days > 1:
            letter = 'D'
            number = 1
            reduced_granularity = letter
            
        elif hours > 1:
            base   = 2
            letter = 'H'
            number = base*round(hours/base)
            if number > 12:
                number = 12
            reduced_granularity = letter + str(number)
                
        elif minutes > 1:
            base   = 15 
            letter = 'M'
            number = base*round(minutes/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
            
        else: 
            base   = 15 
            letter = 'S'
            number = base*round(seconds/base)
            if number > 30:
                number = 30
            reduced_granularity = letter + str(number)
        
        if reduced_granularity[1:] == '0':
            reduced_granularity = reduced_granularity[0] + '1'
        
        return reduced_granularity
    
    def get_NAV(self):
        ''' Returns Net Asset Value of account. '''
        response = self.api.account.get(accountID=self.ACCOUNT_ID)
        return response.body["account"].NAV



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Utility functions for Oanda.
----------------------------

"""

import pandas as pd
from datetime import datetime, timedelta
import numpy as np
from autotrader.brokers.broker_utils import BrokerUtils

class Utils(BrokerUtils):
    def __init__(self):
        return
    
    def update_data_with_candle(self, data, latest_candle):
        ''' Appends the latest candle to the data. '''
        # latest candle will be in the form of a list.
        # [time, Open, High, Low, Close]
        # Need to convert this to a df.
        # latest_candle = ['2021-06-14 02:54:30', 1.20994, 1.20994, 1.20994, 1.20994]
        candle_time = datetime.strptime(latest_candle[0],
                                        '%Y-%m-%d %H:%M:%S')
        
        candle = pd.DataFrame({"Open": latest_candle[1], 
                          "High": latest_candle[2], 
                          "Low": latest_candle[3],
                          "Close": latest_candle[4]},
                          index=[candle_time])
        new_data = pd.concat([data, candle])
        
        return new_data
    
    def last_period(self, current_time, granularity, current_candle = False):
        ''' 
            Returns a datetime object corresponding to the last candle that closed,
            based on the current time and the granularity of the candle.
            
            Note: if the current time is 10:15:01, the instrument will have just 
            begun forming the 10:15 candle. As such, the latest candle in the 
            instruments price history is the 10:00 candle, as that is the last 
            candle that has closed.
            
        '''    
        
        letter_to_unit = {'S': 'second',
                          'M': 'minute',
                          'H': 'hour',
                          'D': 'day'}
        
        letter = granularity[0]
        if len(granularity) > 1:
            number = int(granularity[1:])
        else:
            number = 1
        
        current_period      = getattr(current_time, letter_to_unit[letter])
        if current_candle:
            last_period         = number * np.floor(current_period/number)
        else:
            last_period         = number * np.floor(current_period/number) - number
        
        if letter == 'S':
            td = timedelta(microseconds = current_time.microsecond,
                           seconds = current_time.second - last_period)
        elif letter == 'M':
            td = timedelta(microseconds = current_time.microsecond,
                           seconds = current_time.second,
                           minutes = current_time.minute - last_period)
        elif letter == 'H':
            td = timedelta(microseconds = current_time.microsecond,
                           seconds = current_time.second,
                           minutes = current_time.minute,
                           hours = current_time.hour - last_period)
        elif letter == 'H':
            td = timedelta(microseconds = current_time.microsecond,
                           seconds = current_time.second,
                           minutes = current_time.minute,
                           hours = current_time.hour,
                           days = current_time.day - last_period)
            
        last_candle_closed = current_time - td
        
        return last_candle_closed
    
    
    
    def trade_summary(self, raw_livetrade_summary, ohlc_data, granularity):
        ''' Constructs trade summary dataframe from Oanda trade history. '''
        
        # Need to create a new coloumn with ohlc_data int index
        # and reutrn new trade summary, after using last_period function
        # Will also need to round the trade summary times to the nearest granularity
        
        # Create ohlc_data int index
        modified_data           = ohlc_data
        modified_data['date']   = modified_data.index
        modified_data           = modified_data.reset_index(drop = True)
        modified_data['data_index'] = modified_data.index
        
        # Round trade summary times to nearest candle
        rounded_times           = [self.last_period(dt, 'M15', current_candle=True) \
                                   for dt in pd.to_datetime(raw_livetrade_summary.Date.values, utc=True)]
        
        # Reset trade summary and add rounded times
        raw_livetrade_summary   = raw_livetrade_summary.reset_index()
        raw_livetrade_summary['Date'] = rounded_times
        
        # Merge to preserve data int index
        livetrade_summary       = pd.merge(modified_data, raw_livetrade_summary, left_on='date', right_on='Date')
        
        return livetrade_summary
    
    
    def format_watchlist(self, raw_watchlist):
        
        watchlist = []
        for instrument in raw_watchlist:
            if str(instrument) != 'nan':
                formatted_instrument = instrument[:3] + "_" + instrument[-3:]
                watchlist.append(formatted_instrument)
        
        return watchlist


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 23 10:19:44 2021

@author: kieran
"""

from getopt import getopt
import sys
import os
from autotrader.emailing import send_order_summary
import yaml


def main(uo_dict):
    email_dir   = os.path.dirname(os.path.abspath(__file__))
    file_dir    = os.path.join(email_dir, '../logfiles')
    file_path   = os.path.join(file_dir, uo_dict["filename"])
    
    global_config   = read_yaml(email_dir + '/../config/GLOBAL.yaml')
    host_email      = global_config["EMAILING"]["HOST_ACCOUNT"]
    mailing_list    = global_config["EMAILING"]["MAILING_LIST"]
    
    send_order_summary(file_path, mailing_list, host_email)


def read_yaml(file_path):
    '''Function to read and extract contents from .yaml file.'''
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


short_options = "f:"
long_options = ['file=']

if __name__ == '__main__':
    options, r = getopt(sys.argv[1:], 
                          short_options, 
                          long_options
                          )
    
    # Defaults
    filename = None
    
    for opt, arg in options:
        if opt in ('-f', '--file'):
            filename = arg
        
    uo_dict = {'filename': filename}

    main(uo_dict)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jul  7 12:39:11 2021

@author: kieran
"""

def get_config(environment, global_config, feed):
    '''
        Returns the configuration dictionary based on the requested 
        environment.
        
    '''
    
    if environment == 'real':
        if feed == 'OANDA':
            data_source     = 'OANDA'
            api             = global_config['OANDA']['LIVE_API']
            access_token    = global_config['OANDA']['ACCESS_TOKEN']
            account_id      = global_config['OANDA']['DEFAULT_ACCOUNT_ID']
            port            = global_config['OANDA']['PORT']
            
            config_dict = {'data_source'    : data_source,
                           'API'            : api, 
                           'ACCESS_TOKEN'   : access_token, 
                           'ACCOUNT_ID'     : account_id, 
                           'PORT'           : port}
            
        elif feed == 'IB':
            data_source     = 'IB'
            print("Interactive brokers not supported yet.")
            
        elif feed == 'yahoo':
            data_source     = 'yfinance'
            config_dict = {'data_source'    : data_source}
            
        else:
            print("Unrecognised data feed. Please check config and retry.")
            
    
    else:
        if feed.upper() == 'OANDA':
            data_source     = 'OANDA'
            api             = global_config['OANDA']['PRACTICE_API']
            access_token    = global_config['OANDA']['ACCESS_TOKEN']
            account_id      = global_config['OANDA']['DEFAULT_ACCOUNT_ID']
            port            = global_config['OANDA']['PORT']
            
            config_dict = {'data_source'    : data_source,
                           'API'            : api, 
                           'ACCESS_TOKEN'   : access_token, 
                           'ACCOUNT_ID'     : account_id, 
                           'PORT'           : port}
            
        elif feed.upper() == 'IB':
            data_source     = 'IB'
            print("Interactive brokers not supported yet.")
            return
            
        elif feed.upper() == 'YAHOO':
            data_source = 'yfinance'
            config_dict = {'data_source'    : data_source}
            
        else:
            print("Unrecognised data feed {}.".format(feed) + \
                  "Please check global config and retry.")
    
    
    return config_dict